from __future__ import annotations

import asyncio
from logging import Handler
from pathlib import Path
from typing import Literal

import toml
import typer
from loguru import logger
from rich.highlighter import NullHighlighter
from rich.logging import RichHandler

from nanoflow import WorkflowConfig
from nanoflow.executor import Executor
from nanoflow.utils import layer_nodes

app = typer.Typer()


def init_logger(log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], handler: Handler):
    logger.remove()
    logger.add(handler, format="{message}", level=log_level)


@app.command()
def run(
    config_path: Path,
    *,
    use_tui: bool = False,
    try_run: bool = False,
):
    handler = RichHandler(highlighter=NullHighlighter(), markup=True)
    init_logger("DEBUG", handler)
    workflow_config = WorkflowConfig.model_validate(toml.load(config_path))
    if try_run:
        if use_tui:
            logger.warning("[blue bold]use-tui[/] is ignored when try-run is used")
        layered_nodes = layer_nodes(workflow_config.to_nodes())
        for i, layer in enumerate(layered_nodes):
            logger.info(f"Layer [blue bold]{i}[/]")
            for node in layer:
                print(workflow_config.tasks[node].get_command())
        return
    if use_tui:  # pragma: no cover
        from nanoflow.tui import Nanoflow

        app = Nanoflow(workflow_config)
        executor = Executor.from_configs(workflow_config, update_hook=app.update_log)

        async def start():
            await asyncio.gather(executor.workflow(), app.run_async())

        asyncio.run(start())
    else:
        executor = Executor.from_configs(workflow_config)
        executor.run()


@app.command()
def try_run(config_path: Path):
    run(config_path, try_run=True)



from __future__ import annotations

import asyncio
import subprocess
from abc import abstractmethod
from collections.abc import Hashable, Sequence

from loguru import logger


class ResourcePool[T: Hashable]:
    def __init__(self, resources: Sequence[T]):
        self.resources = {res: asyncio.Semaphore(1) for res in resources}

    async def acquire(self) -> T:
        while True:
            for res, sem in self.resources.items():
                if sem.locked():
                    continue
                if await sem.acquire():
                    return res
            await asyncio.sleep(0.1)

    def release(self, res: T):
        self.resources[res].release()


class DynamicResourcePool[T](ResourcePool[T]):
    resources: dict[T, asyncio.Semaphore]
    used_resources: set[T]

    def __init__(self):
        self.resources = {}
        self.used_resources = set()

    @abstractmethod
    def get_available_resources(self) -> set[T]: ...

    def update(self):
        available_resources = self.get_available_resources()
        for new_res in available_resources - self.resources.keys():
            logger.info(f"Adding new resource {new_res}")
            self.resources[new_res] = asyncio.Semaphore(1)
        for res in self.resources.keys() - available_resources - self.used_resources:
            logger.info(f"Removing resource {res}")
            self.resources.pop(res)

    async def acquire(self) -> T:
        while True:
            self.update()
            for res, sem in self.resources.items():
                if sem.locked():
                    continue
                if await sem.acquire():
                    self.used_resources.add(res)
                    return res
            await asyncio.sleep(0.1)

    def release(self, res: T):
        self.resources[res].release()
        self.used_resources.remove(res)


class GPUResourcePool(DynamicResourcePool[str]):
    def __init__(self, threshold: float = 0.05):
        self.threshold = threshold
        super().__init__()

    def get_available_resources(self) -> set[str]:
        gpu_info: str = subprocess.check_output(
            [
                "nvidia-smi",
                "--query-gpu=index,utilization.gpu,memory.used,memory.total",
                "--format=csv,nounits,noheader",
            ]
        ).decode("utf-8")
        gpus = [tuple(map(int, line.split(","))) for line in gpu_info.strip().split("\n")]

        free_gpus = set()
        for index, gpu_usage_ratio, used_mem, total_mem in gpus:
            mem_usage_ratio = used_mem / total_mem
            if gpu_usage_ratio <= self.threshold and mem_usage_ratio <= self.threshold:
                free_gpus.add(str(index))

        return free_gpus


class UnlimitedPool[T](ResourcePool[T]):
    def __init__(self, resource: T):
        self.resource = resource

    async def acquire(self) -> T:
        return self.resource

    def release(self, res: T):
        pass
